# Amethyst - An experimental Minecraft server written in Elixir.
# Copyright (C) 2024 KodiCraft
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

defmodule Amethyst.ConnectionHandler do
  @moduledoc """
  This module is responsible for handling incoming packets and sending outgoing packets. It keeps track of what state the connection is in and which game should
  receive the packets.
  """
  alias Amethyst.Minecraft.Write
  require Logger

  @spec child_spec(:gen_tcp.socket()) :: Supervisor.child_spec()
  def child_spec(socket) do
    %{
      id: __MODULE__,
      start: {__MODULE__, :start, [socket, Amethyst.ConnectionState.Handshake, 0]}
    }
  end

  @spec start(:gen_tcp.socket(), atom(), integer()) :: no_return()
  def start(socket, connstate, version) do
    {:ok, spawn(fn ->
      Process.set_label("ConnectionHandler for #{inspect(socket)}")
      loop(socket, connstate, version, %{})
    end)}
  end

  @spec start_link(:gen_tcp.socket(), atom(), integer()) :: no_return()
  def start_link(socket, connstate, version) do
    {:ok, spawn_link(fn ->
      Process.set_label("ConnectionHandler for #{inspect(socket)}")
      loop(socket, connstate, version, %{})
    end)}
  end

  @spec loop(:gen_tcp.socket(), atom(), integer(), map()) :: no_return()
  defp loop(socket, connstate, version, state) do
    receive do
      {:disconnect, reason} ->
        disconnect(socket, reason, connstate, version, state)
        Process.exit(self(), :normal)
      {:set_state, newstate} ->
        Logger.debug("Switching to state #{newstate} from #{connstate}")
        loop(socket, newstate, version, state)
      {:set_version, newversion} ->
        Logger.debug("Switching to version #{newversion} from #{version}")
        loop(socket, connstate, newversion, state)
      {:set_encryption, secret} ->
        Logger.debug("Enabling encryption with shared secret #{inspect(secret)}")
        encryption_state = :crypto.crypto_init(:aes_128_cfb8, secret, secret, true)
        decryption_state = :crypto.crypto_init(:aes_128_cfb8, secret, secret, false)
        state = state |> Map.put(:encryption_state, encryption_state) |> Map.put(:decryption_state, decryption_state)
        loop(socket, connstate, version, state)
      {:set_compression, threshold} ->
        Logger.debug("Enabling comrpession with threshold #{threshold}")
        state = Map.put(state, :compression, threshold)
        loop(socket, connstate, version, state)
      {:set_position, position} ->
        prev_position = Map.get(state, :position)
        state = Map.put(state, :position, position)
        # If there was no prev position, we consider that we
        # definitely moved
        prev_cp = if prev_position == nil do nil else chunk_pos(elem(prev_position, 0), elem(prev_position, 2)) end
        cp = chunk_pos(elem(position, 0), elem(position, 2))
        if prev_cp != cp do
          Logger.debug("Client entered new chunk #{inspect(cp)}")
          # We changed chunk borders, update center chunk and begin sending new chunks
          send(self(), {:send_packet, %{
            packet_type: :set_center_chunk,
            chunk_x: elem(cp, 0),
            chunk_z: elem(cp, 1)
          }})
          # Figure out which new chunks are visible
          prev_chunks =
            if prev_cp == nil do
              MapSet.new([])
            else
              MapSet.new(visible_chunks_from(elem(prev_cp, 0), elem(prev_cp, 1), 16)) # This 16 would be the server view distance!!
            end
          chunks = MapSet.new(visible_chunks_from(elem(cp, 0), elem(cp, 1), 16))
          new_chunks = MapSet.difference(chunks, prev_chunks)
          Logger.debug("Sending #{MapSet.size(new_chunks)} chunks...")
          # We can process all chunks in parallel
          me = self()
          ts = state |> Map.get(:game) |> Map.get(:refs) |> Map.get(:task_supervisor)
          Task.Supervisor.async(ts, fn ->
            Task.Supervisor.async_stream(ts,
            new_chunks,
            fn chunk -> process_chunk(me, chunk, state) end,
            [ordered: false]
          ) |> Stream.run() end)
        end
        loop(socket, connstate, version, state)
      {:send_packet, packet} ->
        # Logger.debug("Sending packet #{inspect(packet)}")
        send_packet(socket, connstate, packet, version, state)
        loop(socket, connstate, version, state)
      after 0 ->
        # Received stuff from the connection receiver is lower priority
        receive do
          :closed ->
            Logger.info("Connection #{inspect(socket)} closed.")
            Process.exit(self(), :normal)
          {:get_encryption, from} ->
            send(from, Map.get(state, :decryption_state))
            loop(socket, connstate, version, state)
          {:get_compression, from} ->
            send(from, Map.get(state, :compression))
            loop(socket, connstate, version, state)
          {:packet, id, data} ->
            state = handle_packet(id, data, connstate, version, state)
            loop(socket, connstate, version, state)
        end
    end
  end

  defp chunk_pos(x, z) do
    {floor(round(x) / 16.0), floor(round(z) / 16.0)}
  end

  # x, z here is chunk position
  defp visible_chunks_from(x, z, view_distance) do
    (x - view_distance - 3 .. x + view_distance + 3) |> Enum.flat_map(fn ix ->
      (z - view_distance - 3 .. z + view_distance + 3) |> Enum.map(fn iz ->
        {ix, iz}
      end)
    end)
  end

  defp process_chunk(to, chunk, state) do
    import Amethyst.NBT.Write
    alias Amethyst.Minecraft.Write

    chunk_array = Amethyst.Game.chunk(Map.get(state, :game), chunk)
    {cx, cz} = chunk

    # TODO: Actually do heightmaps
    # TODO: Doing all this processing could be at home somewhere else, as here it's
    # not version-agnostic here
    heightmaps = compound(%{})

    data = Enum.chunk_every(chunk_array, 16, 16, 0) # 0 -> air
      |> Enum.reduce("", fn chunk_section, acc ->
        blocks_and_lights = chunk_section |> List.flatten()
        block_count = blocks_and_lights |> Enum.filter(fn {bs, _, _} -> bs != 0 end) |> length

        # Put together the palette
        unique_blocks = MapSet.new(blocks_and_lights |> Enum.map(fn {bs, _, _} -> bs end))
        min_bpe = MapSet.size(unique_blocks) |> :math.log2() |> ceil()

        paletted_container_data = case min_bpe do
          0 ->
          # SINGLE VALUED
            Write.ubyte(0) <>
            Write.varint(MapSet.to_list(unique_blocks) |> List.first()) <>
            Write.varint(0) # No data, empty pallette

          min_bpe when min_bpe in 1..8 ->
          # INDIRECT
            # Minimum bpe accepted by minecraft is 4
            bpe = max(min_bpe, 4)
            palette = MapSet.to_list(unique_blocks) |>
              Enum.with_index() |>
              Map.new(fn {v, i} -> {v, i} end)
            paletted_blocks = blocks_and_lights |>
              Enum.map(fn {bs, _, _} -> Map.get(palette, bs) end)
            paletted_data = long_aligned_bit_string_reduce(paletted_blocks, bpe,
              fn bs -> <<bs::signed-integer-big-size(bpe)>> end)
            Write.ubyte(bpe) <>
            Write.varint(map_size(palette)) <>
            (Enum.sort(palette, fn {_k1, v1}, {_k2, v2} -> v1 < v2 end) |>
              Enum.reduce("", fn {k, _v}, acc -> acc <> Write.varint(k) end)) <>
            Write.varint(floor(bit_size(paletted_data) / 64)) <>
            paletted_data

          _ ->
          # DIRECT
            data = long_aligned_bit_string_reduce(blocks_and_lights, 15,
              fn {bs, _sl, _bl} -> <<bs::signed-integer-big-size(15)>> end)
            Write.ubyte(15) <>
            Write.varint(floor(bit_size(data) / 64)) <>
            data
        end

        acc <> Write.short(block_count) <> paletted_container_data <>
        <<0::8, 0::8, 0::8>> # TODO: This should be biome data
      end)

  sky_light_sections = Enum.chunk_every(chunk_array, 16, 16, 0)
      |> Enum.map(fn chunk_section ->
        chunk_section |> List.flatten() |> Enum.map(fn {_bs, sl, _bl} -> sl end)
      end)
  non_empty_sky_light_sections = sky_light_sections |> Enum.filter(fn section -> !Enum.all?(section, &(&1 == 0)) end)
  has_sky_light = [false] ++ (sky_light_sections |> Enum.map(fn section -> !Enum.all?(section, &(&1 == 0)) end)) ++ [false]
  empty_sky_light = has_sky_light |> Enum.map(&not/1)
  sky_light_array = non_empty_sky_light_sections |> Enum.map(&(Enum.reduce(&1, <<>>, fn light, acc ->
    <<acc::bitstring, light::size(4)>>
  end)))

  block_light_sections = Enum.chunk_every(chunk_array, 16, 16, 0)
  |> Enum.map(fn chunk_section ->
    chunk_section |> List.flatten() |> Enum.map(fn {_bs, _sl, bl} -> bl end)
  end)
  non_empty_block_light_sections = block_light_sections |> Enum.filter(fn section -> !Enum.all?(section, &(&1 == 0)) end)
  has_block_light = [false] ++ (block_light_sections |> Enum.map(fn section -> !Enum.all?(section, &(&1 == 0)) end)) ++ [false]
  empty_block_light = has_block_light |> Enum.map(&not/1)
  block_light_array = non_empty_block_light_sections |> Enum.map(&(Enum.reduce(&1, <<>>, fn light, acc ->
    <<acc::bitstring, light::size(4)>>
  end)))
  Logger.debug("#{length(block_light_array)} block light sections to send")

  send(to, {:send_packet, %{
    packet_type: :chunk_data_and_update_light,
    chunk_x: cx, chunk_z: cz,
    heightmaps: heightmaps,
    data: data,
    block_entities: [],
    sky_light_mask: has_sky_light,
    block_light_mask: has_block_light,
    empty_sky_light_mask: empty_sky_light,
    empty_block_light_mask: empty_block_light,
    sky_light_arrays: sky_light_array |> Enum.map(&(%{sky_light_array: &1})),
    block_light_arrays: block_light_array |> Enum.map(&(%{block_light_array: &1}))
  }})
  :ok
  end

  defp long_aligned_bit_string_reduce(values, bpe, func) do
    values |> Enum.reduce(<<>>, fn value, acc ->
      ret = func.(value)
      next = <<acc::bitstring, ret::bitstring-size(bpe)>>
      # man i hope they dont suddenly change the size of a long
      if rem(bit_size(next), 64) + bpe > 64 do
        # gotta pad it
        <<next::bitstring, 0::big-size(64 - rem(bit_size(next), 64))>>
      else
        next
      end
    end)
  end

  defp handle_packet(id, data, connstate, version, state) do
    try do
      packet = connstate.deserialize(id, version, data)
      case connstate.handle(packet, version, state) do
        :ok -> state
        {:error, reason} ->
          Logger.error("Error handling packet with ID #{inspect(id, base: :hex)} in state #{connstate}: #{reason}")
          send(self(), {:disconnect, "§cError handling packet #{inspect(id, base: :hex)}:\n#{reason}"})
          state
        newstate ->
          if is_map(newstate) do
            newstate
          else
            Logger.warning("State change to #{inspect(newstate)} is not a map! Did you forget to return :ok?")
            state
          end
      end
    rescue
      e ->
        if Application.get_env(:amethyst, :release, false) do
          send(self(), {:disconnect, "§cError handling packet #{inspect(id, base: :hex)}:\n#{Exception.format(:error, e, __STACKTRACE__)}"})
          Logger.error("Error handling packet with ID #{inspect(id, base: :hex)} in state #{connstate}: #{Exception.format(:error, e, __STACKTRACE__)}")
        else
          Logger.error("Error handling packet with ID #{inspect(id, base: :hex)} in state #{connstate}: #{Exception.format(:error, e, __STACKTRACE__)}")
        end
        state
    end
  end

  defp send_packet(socket, connstate, packet, version, state) do
    try do
      data = connstate.serialize(packet, version)
      container = if Map.get(state, :compression) == nil do
        # Packet ID is included in data
        Write.varint(byte_size(data)) <> data
      else
        threshold = Map.get(state, :compression, 0)
        data_length = byte_size(data)
        if data_length >= threshold do
          compressed = Write.varint(data_length) <> :zlib.compress(data)
          Write.varint(byte_size(compressed)) <> compressed
        else
          compressed = Write.varint(0) <> data
          Write.varint(byte_size(compressed)) <> compressed
        end
      end
      encrypted = if Map.get(state, :encryption_state) == nil do
        container
      else
        Map.get(state, :encryption_state) |> :crypto.crypto_update(container)
      end
      :gen_tcp.send(socket, encrypted)
    rescue
      e ->
        Logger.error("Error sending packet #{inspect(packet)} in state #{connstate}: #{Exception.format(:error, e, __STACKTRACE__)}")
        send(self(), {:disconnect, "§cError sending packet #{inspect(packet)}:\n#{Exception.format(:error, e, __STACKTRACE__)}"})
    end
  end

  defp disconnect(socket, reason, connstate, version, state) do
    Logger.info("Disconnecting connection #{inspect(socket)}")
    Logger.debug("Disconnecting connection #{inspect(socket)}: #{reason}")
    case connstate.disconnect(reason) do
      nil -> nil
      packet -> send_packet(socket, connstate, packet, version, state)
    end
    :gen_tcp.close(socket)
  end
end
